% NavABL.m 
% This program illustrates the use of allocentric bearing to a landmark cells to implement navigation to a series of Way-Points to follow a route.
% Copyright E T Rolls, Oxford Centre for Computational Neuroscience, https://www.oxcns.org. 6 Nov 2019 and 15 April 2020.
% No warranty express or implied is provided: this is research software.
% Further background is available in the following paper:
% Rolls,E.T. (2021) Neurons including spatial view cells, and navigation in primates including humans. Hippocampus.
% If this software is used or adapted for new published research, 
% it is requested that the original publication, which provides details, is cited.

% 'Allocentric bearing to landmark' cells are described by Rolls 2020 Hippocampus 30: 332-353
clear all; close all hidden; format compact; clc

% Set up the database arrays that will define the state space
% The Environment is X=100 x Y=100, with 1,1 the top left, and 100,100 the bottom right, with E=100
% Landmark 1 is at 1,1; L2 is at 100,1; L3 is at 100,100; L4 is at 1, 100.
% The nodes contain the information about the steps in the route to be followed.
% In the brain, the information in the nodes could be read out from a continuous attractor network, or from a short-term memory (Rolls2021).

N = 4; % Number of nodes in the continuous attractor / sequence generator. Node 1 defines the starting point in the state space
L = 4; % number of landmarks
% The following is the information in each node
ABL = ones(N, 1) * -1; % Allocentric Bearing to a landmark used to define when each leg of the navigation ends.
B = ones(N, 1) * -1; % Body turns, 0 to 7 for egocentric whole body turns 45 deg apart. 0 is Forward; 2 is right; 6 is left. -1 is stop.
HD = ones(N, 1) * -1; % Head directions to turn to in allocentric coordinates. Can be 0-7 for compass direction starting with North.
% Only one of B or HD is used. In this program, head direction cells are used, but could be substituted with body turn cells

E = 100; % the size of the environment in which navigation will take place. Default 100
Env = zeros(E, E); % the environment. E(1,1) is top left.

vidfile = VideoWriter('NavABL.mp4','MPEG-4');
open(vidfile);

% Now enter some values using allocentric bearing to landmark cells. These will be used in the sequence generator
nN = 1; % define the environment for Node 1
ABL(nN) = 45; % the bearing to a landmark at which a turn must be made
B(nN) = 0; % Body turn = 0
HD(nN) = 2; % East

nN = 2; % define the environment for Node 2
ABL(nN) = 135; % the bearing to a landmark at which a turn must be made
B(nN) = 2; % Body turn = Right 90 deg
HD(nN) = 4; % South

nN = 3; % define the environment for Node 3
ABL(nN) = 225; % the bearing to a landmark at which a turn must be made
B(nN) = 2; % Body turn = Right 90 deg
HD(nN) = 6; % West

nN = 4; % define the environment for Node 4
ABL(nN) = 315; % the bearing to a landmark at which a turn must be made
B(nN) = 0; % No body turn
HD(nN) = 6; % West

% the following are just for the labels for the display
WayPointX(1) = 28; % the start X position in Env.
WayPointY(1)= 33; % the start Y position in Env
WayPointX(2) = 73;
WayPointY(2) = 30;
WayPointX(3) = 73;
WayPointY(3) = 73;
WayPointX(4) = 20;
WayPointY(4) = 77;

% Now try some navigation
X = 30; % the start X position in Env
Y = 30; % the start Y position in Env
figure(1);
imagesc(Env);

for nN = 1 : N - 1 % work through the nodes, as a continuous attractor would
    H = HD(nN); % recall the head direction from the attractor net for this leg of the navigation
    for mov = 1 : 100 % keep moving until the correct bearing to a landmark is reached
        BearingL1 = atand((1-X) / Y); % calculate the bearings to all landmarks to show how, even though only one is needed for each leg
        if BearingL1 < 0
            BearingL1 = 360 + BearingL1;
        end
        BearingL2 = atand((100-X) / Y);
        if BearingL2 < 0
            BearingL2 = 360 + BearingL2;
        end
        BearingL3 = 180 - atand((100-X) / (100-Y));
        if BearingL3 < 0
            BearingL3 = 360 + BearingL3;
        end
        BearingL4 = 180 + atand((X) / (100-Y));
        if BearingL4 < 0
            BearingL4 = 360 + BearingL4;
        end
         
        % Next implement navigation for each leg, and turn when a given bearing to a landmark is reached.
        if nN == 1 % H will have been recalled from the sequence generator
            X = X + 1; % implement locomotion in the direction specified by H, East
            if BearingL2 < ABL(nN) 
                fprintf('Bearing of L2 is <  45, turning South\n');
                break
            end
        elseif nN == 2 % H will have been recalled from the sequence generator
            Y = Y + 1; % implement locomotion in the direction specified by H, South
            if BearingL3 < ABL(nN) 
                fprintf('Bearing of L3 is < 135, turning West\n');
                break
            end
        elseif nN == 3 % H will have been recalled from the sequence generator
            X = X - 1; % implement locomotion in the direction specified by H, West
            if BearingL4 < ABL(nN) 
                fprintf('Bearing of L3 is < 225, turning West\n');
                break
            end           
    
        end

        if X < 2 X=2; end % check that movements are within the environment and visible on the display
        if X > E-1 X=E-1; end
        if Y < 2 Y=2; end
        if Y > E-1 Y=E-1; end

        %fprintf('X=%3d  Y=%3d  BearingL1=%3.0f  L2=%3.0f  L3=%3.0f  L4=%3.0f\n', X, Y, BearingL1, BearingL2, BearingL3, BearingL4);
        Env(Y, X) = 1;  % add the current place to the animation 
        imagesc(Env);
        title('Navigation with allocentric bearing to landmark Cells NavABL.m');
        xlabel('X'); ylabel('Y');
        h=text(2,3,'L1','Fontsize', 14);
        h.Color=[1 0 0]; % red
        h=text(96,3,'L2','Fontsize', 14);
        h.Color=[1 0 0]; % red
        h=text(96,97,'L3','Fontsize', 14);
        h.Color=[1 0 0]; % red
        %  h=text(97,97,'L3', 'Color', [1, 0, 0]); % for Octave
        h=text(2,97,'L4','Fontsize', 14);
        h.Color=[1 0 0]; % red
%         h=text(StartX,StartY,'Start');
%         h.Color=[0 1 0]; % green
        h=text(WayPointX(1)+1,WayPointY(1),'W1','Fontsize', 14);
        h.Color=[0 1 0]; % green
        h=text(WayPointX(2)+1,WayPointY(2),'W2','Fontsize', 14);
        h.Color=[0 1 0]; % green
        h=text(WayPointX(3)+1,WayPointY(3),'W3','Fontsize', 14);
        h.Color=[0 1 0]; % green
        h=text(WayPointX(4)+1,WayPointY(4),'W4','Fontsize', 14);
        h.Color=[0 1 0]; % green
        frame = getframe(gcf);
        writeVideo(vidfile, frame);
        pause(0.02); % watch the agent move
    end
end
close(vidfile)

    
    
