% NavSVC.m 
% This program illustrates the use of spatial view cells to implement navigation to a series of landmarks to follow a route.
% Copyright E T Rolls, Oxford Centre for Computational Neuroscience, https://www.oxcns.org. 8 Nov 2019, 15 April 2020 and 28 February 2021.
% No warranty express or implied is provided: this is research software.
% If this software is used or adapted for new published research, 
% it is requested that the original publication, which provides details, is cited.
% If the software is used with Octave, please comment out the lines to create a video using 'vidfile'
% Further background is available in the following paper and book:
% Rolls,E.T. (2021) Neurons including hippocampal spatial view cells, and navigation in primates including humans. Hippocampus.
% Rolls,E.T. (2021) Brain Computations: What and How. Oxford University Press.

clear all; close all hidden; format compact; clc
% Set up the database arrays that will define the state space
% The Environment is X=100 x Y=100, with 1,1 the top left, and 100,100 the bottom right. (For E = 100)
% The nodes contain the information about the steps in the route to be followed.
% In the brain, the information in the nodes could be read out from a continuous attractor network, or from a short-term memory. 

N = 3; % Number of nodes in the sequence generator, which could be a continuous attractor network (Rolls 2021 Brain Computations. Oxford University Press). 
       % Each node is for a different leg of the navigational trajectory
E = 100; % the size of the environment in which navigation will take place. Default 100
Env = zeros(E, E); % the environment. E(1,1) is top left.
NavDir = 0;   % Navigational Direction. The individual corrects this depending on where the landmark detected by the spatial view cell is relative to the current NavDir.
k = 0.2;      % The proportion by which NavDir is corrected towards the direction of the next landmark detected by a spatial view cell.
              % This is to provide similarity to what might happen with a taxis: a correction is made if the agent is not heading towards the landmark detected by a spatial view cell.
L1X = 10;     % the X coordinate for L1, landmark 1
L1Y = 10;     % the Y coordinate for L1
L2X = 90;     % the X coordinate for L2
L2Y = 10;     % the Y coordinate for L2
L3X = 90;     % the X coordinate for L3
L3Y = 90;     % the Y coordinate for L3
L4X = 10;     % the X coordinate for L4
L4Y = 90;     % the Y coordinate for L4

vidfile = VideoWriter('NavSVC.mp4','MPEG-4');
open(vidfile);
 
% Now try some navigation. Set the X, Y start position to wherever you like within the range 1,1 to E,E.
X = L1X; % the start X position in Env. Default L1X
Y = L1Y; % the start Y position in Env. Default L1Y
figure(1);
imagesc(Env);

for nN = 1 : N     % work through the nodes, as a continuous attractor or working memory would (see Rolls 2021).
                   % Each node specifies which spatial view cells to select to move towards for that leg
    for mov = 1 : 2 * E % this moves the agent, and the leg stops when the agent becomes close to the next landmark.
        if nN == 1 % This is for the first leg, navigation from the start (default L1) to L2. 
                   % The agent now selects and looks for landmark 2 (using spatial view cells) and just moves towards it using taxis.
                   % The agent does not need to calculate the place where it is located,
                   % only that the agent should move towards L2.
                   % This could be implemented by looking at L2, and moving towards where one is looking using taxis.
                   % One way to implement this is as follows for computational convenience, but other methods of approach to where one is looking, that is, taxis, are possible.
            SpatialViewDirection = SpatialViewDir(X, L2X, Y, L2Y); % This gives the direction in degrees to the next landmark, the part of the environment where the spatial view cell fires
                   % This angle is used just to enable the agent to move towards the landmark for computational convenience.
                   % Other methods could be used to implement taxis, that is, approach to what is being viewed and is activating spatial view cells.
                   % For example, the individual could rotate its head direction to the angle that made the spatial view cell
                   % for the next landmark fire maximally with the eyes looking straight ahead, and move in that head direction.
            NavDir = NavDir + k * (SpatialViewDirection - NavDir); % correct NavDir depending on whether the landmark is left or right of the current heading (NavDir)
            fprintf('Leg 1  X=%3.0f  Y=%3.0f Navigational Direction towards L2=%3.0f degrees  Landmark Bearing of L2=%3.0f deg\n', X, Y, NavDir, SpatialViewDirection);
            if abs(L2X - X) < 3 && abs(L2Y - Y) < 3 % if we are close to the correct landmark, break and look for the next landmark
                break
            end
        elseif nN == 2 % select Landmark 3 and move towards it.
            SpatialViewDirection = SpatialViewDir(X, L3X, Y, L3Y);
            % There could be additional body turn or head direction information
            % to help the agent find the landmark at the start of the leg, 
            % as part of the instructions, but this is not necessary.
            NavDir = NavDir + k * (SpatialViewDirection - NavDir); % correct NavDir depending on the spatial view cells
            fprintf('Leg 2  X=%3.0f  Y=%3.0f Navigational Direction towards L3=%3.0f degrees  Landmark Bearing of L3=%3.0f deg\n', X, Y, NavDir, SpatialViewDirection);
            if abs(L3X - X) < 3 && abs(L3Y - Y) < 3 % if we are close to the correct landmark, look for the next
                break
            end
        elseif nN == 3 % select Landmark 4 using spatial view cells and move towards it.
            SpatialViewDirection = SpatialViewDir(X, L4X, Y, L4Y);
            NavDir = NavDir + k * (SpatialViewDirection - NavDir); % correct NavDir depending on depending on the spatial view cells
            fprintf('Leg 3  X=%3.0f  Y=%3.0f Navigational Direction towards L4=%3.0f degrees  Landmark Bearing of L4=%3.0f deg\n', X, Y, NavDir, SpatialViewDirection);
            if abs(L4X - X) < 5 && abs(L4Y - Y) < 5 % if we are close to the correct landmark, look for the next
                fprintf('Finished successfully at X=%3.0f  Y=%3.0f\n', X, Y);
                break
            end
        end
        
        % Now that the spatial view cell has found where the next landmark is, display where the agent is looking on the screen.
        % Calculate where the eyes are looking
        EyeX = X + 10 * sind(SpatialViewDirection);
        EyeY = Y - 10 * cosd(SpatialViewDirection);
        line([X, EyeX], [Y, EyeY],'Color','red', 'LineWidth', 1); % draw a red line to show where the eyes are looking to make the spatial view cell fire
        pause(0.04);

        
        % Next, the following moves the agent towards the Landmark that has NavDir as its bearing.
        X = X + sind(NavDir);
        Y = Y - cosd(NavDir);

        if X < 2 X=2; end % check that movements are within the environment and visible on the display
        if X > E-1 X=E-1; end
        if Y < 2 Y=2; end
        if Y > E-1 Y=E-1; end
        Xpos = floor(X); % show the progress on the display. Make sure that we have integers
        Ypos = floor(Y);
        Env(Ypos, Xpos) = 1; % add the current place to the animation
        imagesc(Env);
        line([X, EyeX], [Y, EyeY],'Color','red', 'LineWidth', 1); % repeat this to include it in the video.
        title('Navigation with spatial view cells to Landmarks 2 to 3 to 4 NavSVC.m');
        xlabel('X'); ylabel('Y');
        h=text(L1X,L1Y,'L1','Fontsize', 14); % place text on the animation
        h.Color=[1 0 0]; % red
        h=text(L2X,L2Y,'L2','Fontsize', 14);
        h.Color=[1 0 0]; % red
        h=text(L3X,L3Y,'L3','Fontsize', 14);
        h.Color=[1 0 0]; % red
        h=text(L4X,L4Y,'L4','Fontsize', 14);
        h.Color=[1 0 0]; % red
        frame = getframe(gcf);
        writeVideo(vidfile, frame);
        %pause(0.02); % watch the agent move
    end
end
close(vidfile)

function Angle = SpatialViewDir(X, LX, Y, LY)           % Calculate the direction from the agent's X,Y position to a landmark at LX,LY that activates a spatial view cell. North is 0 degrees.
    if X == LX; X = X + 1; end   % trap 0
    if Y == LY; Y = Y + 1; end   % trap 0
    if X < LX && Y < LY
        Angle = 90 + atand((LY - Y) / (LX - X));        % This is the angle that makes the spatial view cell fire for leg 1. 
    elseif X > LX  && Y < LY
        Angle  = 270 - atand((LY - Y) / (X - LX));
    elseif X > LX  && Y > LY
        Angle  = 360 - atand((X - LX) / (Y - LY));
    elseif X < LX  && Y > LY
        Angle  = atand((LX - X) / (Y - LY));
    end  
end
        
    
    
