% NavTRI.m 
% This program illustrates the use of triangulation using allocentric bearing to a landmark cells to implement navigation to a series of Way-Points to follow a route.
% Full 'seafaring' navigation is implemented using a series of WayPoints.
% Bearing to landmarks using 'allocentric bearing to landmark' cells are used to triangulate the place of the agent, 
% and then bearings are computed to move the agent towards the next WayPoint.
% Copyright E T Rolls, Oxford Centre for Computational Neuroscience, https://www.oxcns.org. 25 Nov 2019 and 15 April 2020.
% No warranty express or implied is provided: this is research software.
% If this software is used or adapted for new published research, 
% it is requested that the original publication, which provides details, is cited.
% See Rolls,E.T. (2021) Brain Computations: What and How. Oxford University Press.
% Further background is available in the following paper:
% Rolls,E.T. (2021) Neurons including spatial view cells, and navigation in primates including humans. Hippocampus.
%
% The function triangulationToTal.m is called
clear all; close all hidden; format compact; clc

% Set up the database arrays that will define the state space
% The Environment is X=100 x Y=100, with 1,1 the top left, and 100,100 the bottom right
% Landmark 1 is at 1,1; L2 is at 100,1; L3 is at 100,100; L4 is at 1, 100.
% The nodes contain the information about the steps in the route to be followed.
% In the brain, the information in the nodes could be read out from a continuous attractor network, or from a short-term memory (Rolls2021).

N = 4; % Number of nodes in the continuous attractor. Node 1 defines the starting state space
L = 4; % number of landmarks
WayPointX = ones(N, 1); % The WayPoint X value for each node
WayPointY = ones(N, 1);
E = 100; % the size of the environment in which navigation will take place
Env = zeros(E, E); % the environment. E(1,1) is top left.

vidfile = VideoWriter('NavTRI.mp4','MPEG-4');
open(vidfile);

% Now enter some values by defining the WayPointX and Y for each Node.
% These can be edited to make arbitrary paths within the X,Y space of 1-100
% define the WayPointX, Y values for each node
WayPointX(1) = 10;
WayPointY(1) = 10;
WayPointX(2) = 40;
WayPointY(2) = 90;
WayPointX(3) = 80;
WayPointY(3) = 80;
WayPointX(4) = 90;
WayPointY(4) = 10;

% Now try some navigation
StartX = 30; % the start X position in Env
StartY = 30; % the start Y position in Env
X = StartX;
Y = StartY;
% mark the  WayPoints on the spatial environment
Env(WayPointY(1), WayPointX(1)) = 0.5;
Env(WayPointY(2), WayPointX(2)) = 0.5;
Env(WayPointY(3), WayPointX(3)) = 0.5;
Env(WayPointY(4), WayPointX(4)) = 0.5;
figure(1);
imagesc(Env);

for nN = 1 : N % work through the nodes, as a continuous attractor would
    fprintf('Leg %2d, navigating to X=%3d  Y=%3d\n', nN, WayPointX(nN), WayPointY(nN));
    WayPtX = WayPointX(nN); % The next WayPoint to move is recalled from the continuous attractor net
    WayPtY = WayPointY(nN);
    for mov = 1 : 200 % to make the agent move until the end of this leg is reached when the triangulation shows that the agent is at the next WayPoint
        % first calculate where the agent is by using the bearings to each landmark provided by allocantric bearing to landmark cells, and the triangulation
        BearingL1 = atand((1-X) / Y);
        if BearingL1 < 0
            BearingL1 = 360 + BearingL1;
        end
        BearingL2 = atand((100-X) / Y);
        if BearingL2 < 0
            BearingL2 = 360 + BearingL2;
        end
        BearingL3 = 180 - atand((100-X) / (100-Y));
        if BearingL3 < 0
            BearingL3 = 360 + BearingL3;
        end
        BearingL4 = 180 + atand((X) / (100-Y));
        if BearingL4 < 0
            BearingL4 = 360 + BearingL4;
        end
        % now triangulate
        %[ X Y Q ] = triangulationToTal( A1 , A2 , A3 , x1 , y1 , x2 , y2 , x3 , y3 ); % This function estimates the X, Y position from the bearings to the landmarks:
        [ Xest Yest Q ] = triangulationToTal( BearingL1 * pi/180 , BearingL2 * pi/180 , BearingL3 * pi/180 , 1 , 1 , 100 , 1 , 100 , 100);
        % We can also calculate the distances to the landmarks, in case they are encoded by allocentric bearing to a landmark cells
        DistL1= sqrt(X^2 + Y^2); % Pythagoras
        DistL2 = sqrt((100-X)^2 + Y^2);
        DistL3 = sqrt((100-X)^2 + (100-Y)^2);
        DistL4 = sqrt(X^2 + (100-Y)^2);
        %fprintf('X=%4.0f  Y=%4.0f  Xest=%4.0f  Yest=%4.0f  DistL1=%3.0f  DistL2=%3.0f  DistL3=%3.0f  DistL4=%3.0f\n', X, Y, Xest, Yest, DistL1, DistL2, DistL3, DistL4);
        
        % Now calculate the bearing to the next WayPoint to provide the navigation direction
        if (WayPtX > Xest) && (WayPtY < Yest) % angle 0 to 90 degrees
            NavDir = atand((WayPtX - Xest) / (Yest - WayPtY)); 
        elseif (WayPtX > Xest) && (WayPtY > Yest) % angle 90 to 180
            NavDir = 180 - atand((WayPtX - Xest) / (WayPtY - Yest));
        elseif (WayPtX < Xest) && (WayPtY > Yest) % angle 180 to 270
            NavDir = atand((Xest - WayPtX) / (WayPtY - Yest)) + 180;
        elseif (WayPtX < Xest) && (WayPtY < Yest) % angle 270 to 360   
            NavDir = atand((WayPtX - Xest) / (Yest - WayPtY));
        end
        if NavDir < 0
            NavDir = 360 + NavDir;
        end
        
        % Now move towards the next WayPoint
        X = X + sind(NavDir);
        Y = Y - cosd(NavDir);
        
        if X < 1 X=1; end  % check that movements are within the environment
        if X > E X=E; end
        if Y < 1 Y=1; end
        if Y > E Y=E; end
        fprintf('X=%3.0f  Y=%3.0f  BearingL1=%3.0f  L2=%3.0f  L3=%3.0f  L4=%3.0f  Xest=%4.1f  Yest=%4.1f  WayPtX=%3d  WayPtY=%3d  NavDir=%4.0f\n', X, Y, BearingL1, BearingL2, BearingL3, BearingL4, Xest, Yest, WayPtX, WayPtY, NavDir);
        Xpos = floor(X); % make sure that we have integers
        Ypos = floor(Y);
        Env(Ypos, Xpos) = 1; % add the current place to the animation 
        imagesc(Env);
        title('Navigation with Bearings to Landmarks and Triangulation NavTRI.m');
        xlabel('X'); ylabel('Y');
        h=text(2,3,'L1');
        h.Color=[1 0 0]; % red
        h=text(97,3,'L2');
        h.Color=[1 0 0]; % red
        h=text(97,97,'L3');
        h.Color=[1 0 0]; % red
        h=text(2,97,'L4');
        h.Color=[1 0 0]; % red
        h=text(StartX,StartY,'Start');
        h.Color=[0 1 0]; % green
        h=text(WayPointX(1)+1,WayPointY(1),'W1');
        h.Color=[0 1 0]; % green
        h=text(WayPointX(2)+1,WayPointY(2),'W2');
        h.Color=[0 1 0]; % green
        h=text(WayPointX(3)+1,WayPointY(3),'W3');
        h.Color=[0 1 0]; % green
        h=text(WayPointX(4)+1,WayPointY(4),'W4');
        h.Color=[0 1 0]; % green
        frame = getframe(gcf);
        writeVideo(vidfile, frame);

        pause(0.02); % watch the agent move
        % if we have reached the WayPoint, break and start for the next WayPoint specified in the node sequence memory
        if (abs(Xpos - WayPointX(nN)) <= 2) && (abs(Ypos - WayPointY(nN)) <= 2)
            break;
        end
        %pause
    end
end
close(vidfile)

    
    
